package ColorConverter;

import ClientServer::*;
import GetPut::*;
import FIFO::*;

typedef Bit#(8) Color;
typedef Bit#(8) GrayScale;

typedef struct {
    Color r;
    Color g;
    Color b;
    } RGB deriving(Bits, Eq, FShow);
    
module mkGray(Server#(RGB, GrayScale));
        FIFO#(GrayScale) outputValue <- mkFIFO;
        FIFO#(RGB) inputValue <- mkFIFO;
                
        function GrayScale toGray(RGB rgb);
            UInt#(16) scale = 0;
            UInt#(16) r = extend(unpack(rgb.r)) << 8;
            UInt#(16) g = extend(unpack(rgb.g)) << 8;
            UInt#(16) b = extend(unpack(rgb.b)) << 8;
            
            UInt#(16) factorR = 76;  // floor(0.299 * (1 << 8))
            UInt#(16) factorG = 150; // floor(0.587 * (1 << 8))
            UInt#(16) factorB = 29;  // floor(0.114 * (1 << 8))
            
            // Multiply color with factors and normalize
            UInt#(24) foo = extend(r) * extend(factorR); // Enough space for multiplcation result
            r = truncate(foo >> 8);
            foo = extend(g) * extend(factorG);
            g = truncate(foo >> 8);
            foo = extend(b) * extend(factorB);
            b = truncate(foo >> 8);

            scale = r + g + b;
            
            return pack(scale)[15:8];
        endfunction
                
        rule calc;
            let color = inputValue.first; inputValue.deq;
            let gray = toGray(color);

            outputValue.enq(gray);
        endrule
        
        interface Put request = toPut(inputValue);
        interface Get response = toGet(outputValue);
endmodule

endpackage