package ESABus;

    import GetPut :: *;
    import FIFO :: *;
    import Connectable :: *;

    ///////////////////////////////////////
    // Write Direction
    ///////////////////////////////////////

    // Master

    interface ESABusWr_Master_Fabric;
        method ActionValue#(Tuple2#(Bit#(16), Bit#(8))) request;
    endinterface

    interface ESABusWr_Master;
        interface ESABusWr_Master_Fabric fab;

        interface Put#(Tuple2#(Bit#(16), Bit#(8))) request;
    endinterface

    module mkESABusWr_Master(ESABusWr_Master);

        FIFO#(Tuple2#(Bit#(16), Bit#(8))) requestIn <- mkFIFO();

        interface ESABusWr_Master_Fabric fab;
            method ActionValue#(Tuple2#(Bit#(16), Bit#(8))) request;
                requestIn.deq();
                return requestIn.first();
            endmethod
        endinterface

        interface Put request = toPut(requestIn);
    endmodule

    // Slave

    interface ESABusWr_Slave_Fabric;
        method Action write(Bit#(16) addr, Bit#(8) data);
    endinterface

    interface ESABusWr_Slave;
        interface ESABusWr_Slave_Fabric fab;

        interface Get#(Tuple2#(Bit#(16), Bit#(8))) request;
    endinterface

    module mkESABusWr_Slave(ESABusWr_Slave);
        FIFO#(Tuple2#(Bit#(16), Bit#(8))) requestIn <- mkFIFO();        

        interface ESABusWr_Slave_Fabric fab;
            method Action write(Bit#(16) addr, Bit#(8) data);
                requestIn.enq(tuple2(addr, data));
            endmethod
        endinterface

        interface Get request = toGet(requestIn);
    endmodule

    // Connection
    instance Connectable#(ESABusWr_Master_Fabric, ESABusWr_Slave_Fabric);
        module mkConnection#(ESABusWr_Master_Fabric a, ESABusWr_Slave_Fabric b)(Empty);
            rule forwardWrite;
                let d <- a.request();
                b.write(tpl_1(d), tpl_2(d));
            endrule
        endmodule
    endinstance

    ///////////////////////////////////////
    // Read Direction
    ///////////////////////////////////////

    // Master

    interface ESABusRd_Master_Fabric;
        method ActionValue#(Bit#(16)) addr;
        method Action data(Bit#(8) d);
    endinterface

    interface ESABusRd_Master;
        interface ESABusRd_Master_Fabric fab;

        interface Put#(Bit#(16)) request;
        interface Get#(Bit#(8)) response;
    endinterface

    module mkESABusRd_Master(ESABusRd_Master);

        FIFO#(Bit#(16)) requestIn <- mkFIFO();
        FIFO#(Bit#(8)) responseOut <- mkFIFO();

        interface ESABusRd_Master_Fabric fab;
            method ActionValue#(Bit#(16)) addr;
                requestIn.deq();
                return requestIn.first();
            endmethod

            method Action data(Bit#(8) d);
                responseOut.enq(d);
            endmethod
        endinterface

        interface Put request = toPut(requestIn);
        interface Get response = toGet(responseOut);
    endmodule

    // Slave

    interface ESABusRd_Slave_Fabric;
        method Action addr(Bit#(16) addr);
        method ActionValue#(Bit#(8)) data();
    endinterface

    interface ESABusRd_Slave;
        interface ESABusRd_Slave_Fabric fab;

        interface Get#(Bit#(16)) request;
        interface Put#(Bit#(8)) response;
    endinterface

    module mkESABusRd_Slave(ESABusRd_Slave);
        FIFO#(Bit#(16)) requestIn <- mkFIFO();
        FIFO#(Bit#(8)) responseOut <- mkFIFO();

        interface ESABusRd_Slave_Fabric fab;
            method Action addr(Bit#(16) a);
                requestIn.enq(a);
            endmethod

            method ActionValue#(Bit#(8)) data();
                responseOut.deq();
                return responseOut.first();
            endmethod
        endinterface

        interface Get request = toGet(requestIn);
        interface Put response = toPut(responseOut);
    endmodule

    // Connection
    instance Connectable#(ESABusRd_Master_Fabric, ESABusRd_Slave_Fabric);
        module mkConnection#(ESABusRd_Master_Fabric a, ESABusRd_Slave_Fabric b)(Empty);
            rule forwardRequest;
                let d <- a.addr();
                b.addr(d);
            endrule

            rule forwardResponse;
                let d <- b.data();
                a.data(d);
            endrule
        endmodule
    endinstance

endpackage