package ESAConnect;

    import GetPut :: *;
    import Vector :: *;
    import Arbiter :: *;

    import ESABus :: *;
    import ESAMem :: *;

    module mkESAWrConnector#(Vector#(nMaster, ESABusWr_Master_Fabric) master, Vector#(nSlaves, ESABusWr_Slave_Fabric) slaves)(Empty)
                provisos(Log#(nSlaves, slaveBits));
        Arbiter_IFC#(nMaster) arbiter <- mkArbiter(False);

        Rules sendRequestRules = emptyRules();

        for(Integer i = 0; i < valueOf(nMaster); i = i + 1) begin
            Wire#(Tuple2#(Bit#(16), Bit#(8))) request <- mkWire();
            Wire#(Bit#(slaveBits)) activeSlave <- mkWire();

            rule handleMaster;
                let r <- master[i].request();
                arbiter.clients[i].request();
                request <= r;

                Bit#(slaveBits) slaveSel = 0;
                if(valueOf(nSlaves) != 1)
                    slaveSel = tpl_1(r)[15:16-valueOf(slaveBits)];
                activeSlave <= slaveSel;

                //$display("Fetched request of master %d", i);
            endrule

            for(Integer j = 0; j < valueOf(nSlaves); j = j + 1) begin
                sendRequestRules = rJoinMutuallyExclusive(rules
                    rule sendRequest if(arbiter.clients[i].grant() && (fromInteger(j) == activeSlave || valueOf(nSlaves) == 1));
                        slaves[j].write(tpl_1(request), tpl_2(request));
                        //$display("Master %d sending request to slave %d %x %x", i, activeSlave, tpl_1(request), tpl_2(request));
                    endrule
                endrules, sendRequestRules);
            end
        end

        addRules(sendRequestRules);
    endmodule

    module mkESARdConnector#(Vector#(nMaster, ESABusRd_Master_Fabric) master, Vector#(nSlaves, ESABusRd_Slave_Fabric) slaves)(Empty)
                provisos(Log#(nSlaves, slaveBits));

        Arbiter_IFC#(nMaster) arbiter <- mkArbiter(False);

        Reg#(Bool) transferActive <- mkReg(False);

        Rules sendRequestRules = emptyRules();

        for(Integer i = 0; i < valueOf(nMaster); i = i + 1) begin
            Wire#(Bit#(16)) request <- mkWire();
            Reg#(Bit#(slaveBits)) activeSlave[2] <- mkCReg(2, 0);

            Reg#(Bool) iAmActive <- mkReg(False);

            rule handleMaster if(!transferActive);
                let r <- master[i].addr();
                arbiter.clients[i].request();
                request <= r;

                Bit#(slaveBits) slaveSel = 0;
                if(valueOf(nSlaves) != 1)
                    slaveSel = r[15:16-valueOf(slaveBits)];
                activeSlave[0] <= slaveSel;
                //$display("Fetched request of master %d", i);
            endrule

            for(Integer j = 0; j < valueOf(nSlaves); j = j + 1) begin
                sendRequestRules = rJoinMutuallyExclusive(rules
                    rule sendRequest if(!transferActive && arbiter.clients[i].grant() && (fromInteger(j) == activeSlave[1] || valueOf(nSlaves) == 1));
                        
                        slaves[j].addr(request);

                        //$display("Master %d sending request to slave %d %x", i, activeSlave[1], request);
                        transferActive <= True;
                        iAmActive <= True;
                    endrule

                    rule fetchResponse if(transferActive && iAmActive && (fromInteger(j) == activeSlave[0] || valueOf(nSlaves) == 1));
                        transferActive <= False;
                        iAmActive <= False;
                        let d <- slaves[j].data();
                        master[i].data(d);
                        //$display("Master %d got response %d", i, d);
                    endrule
                endrules, sendRequestRules);
            end
        end

        addRules(sendRequestRules);
    endmodule

    import StmtFSM :: *;

    module mkMultiMasterTest(Empty);

        Vector#(2, ESABusWr_Master) master_wr <- replicateM(mkESABusWr_Master);
        Vector#(2, ESABusRd_Master) master_rd <- replicateM(mkESABusRd_Master);
        Vector#(2, ESAMem) slave <- replicateM(mkESAMem);

        Vector#(2, ESABusWr_Master_Fabric) masterIfc_Wr;
        Vector#(2, ESABusWr_Slave_Fabric) slaveIfc_Wr;

        Vector#(2, ESABusRd_Master_Fabric) masterIfc_Rd;
        Vector#(2, ESABusRd_Slave_Fabric) slaveIfc_Rd;

        for(Integer i = 0; i < 2; i = i + 1) begin
            masterIfc_Wr[i] = master_wr[i].fab;
            masterIfc_Rd[i] = master_rd[i].fab;
            slaveIfc_Wr[i] = slave[i].s_wr;
            slaveIfc_Rd[i] = slave[i].s_rd;
        end

        mkESAWrConnector(masterIfc_Wr, slaveIfc_Wr);
        mkESARdConnector(masterIfc_Rd, slaveIfc_Rd);

        Stmt fsm = {
            seq
                $display("Starting test");
                master_wr[0].request.put(tuple2(16'hAAAB, 8'hCC));
                master_wr[1].request.put(tuple2(16'hAAAB, 8'hDD));
                master_rd[0].request.put(16'hAAAB);
                action
                    let x <- master_rd[0].response.get();
                    $display("Got response %x", x);
                endaction
                master_wr[0].request.put(tuple2(16'h0AAB, 8'hCC));
                master_wr[1].request.put(tuple2(16'hAAAB, 8'hCC));
                delay(1000);
            endseq
        };

        mkAutoFSM(fsm);
    endmodule

    module mkSingleSlaveTest(Empty);

        Vector#(2, ESABusWr_Master) master_wr <- replicateM(mkESABusWr_Master);
        Vector#(2, ESABusRd_Master) master_rd <- replicateM(mkESABusRd_Master);
        Vector#(1, ESAMem) slave <- replicateM(mkESAMem);

        Vector#(2, ESABusWr_Master_Fabric) masterIfc_Wr;
        Vector#(1, ESABusWr_Slave_Fabric) slaveIfc_Wr;

        Vector#(2, ESABusRd_Master_Fabric) masterIfc_Rd;
        Vector#(1, ESABusRd_Slave_Fabric) slaveIfc_Rd;

        for(Integer i = 0; i < 2; i = i + 1) begin
            masterIfc_Wr[i] = master_wr[i].fab;
            masterIfc_Rd[i] = master_rd[i].fab;
        end

        for(Integer i = 0; i < 1; i = i + 1) begin
            slaveIfc_Wr[i] = slave[i].s_wr;
            slaveIfc_Rd[i] = slave[i].s_rd;
        end

        mkESAWrConnector(masterIfc_Wr, slaveIfc_Wr);
        mkESARdConnector(masterIfc_Rd, slaveIfc_Rd);

        Stmt fsm = {
            seq
                $display("Starting test");
                master_wr[0].request.put(tuple2(16'hAAAB, 8'hCC));
                master_wr[1].request.put(tuple2(16'hAAAB, 8'hDD));
                master_rd[0].request.put(16'hAAAB);
                action
                    let x <- master_rd[0].response.get();
                    $display("Got response %x", x);
                endaction
                master_wr[0].request.put(tuple2(16'h0AAB, 8'hCC));
                master_wr[1].request.put(tuple2(16'hAAAB, 8'hCC));
                delay(1000);
            endseq
        };

        mkAutoFSM(fsm);
    endmodule

endpackage