package ESAMem;

    import BlueCheck :: *;
    import FIFO :: *;
    import GetPut :: *;
    import Connectable :: *;
    import RegFile :: *;
    import Clocks :: *;
    import StmtFSM :: *;

    import ESABus :: *;

    interface ESAMem;
        interface ESABusWr_Slave_Fabric s_wr;
        interface ESABusRd_Slave_Fabric s_rd;
    endinterface

    module mkESAMem(ESAMem);
        ESABusWr_Slave writeSlave <- mkESABusWr_Slave();
        ESABusRd_Slave readSlave <- mkESABusRd_Slave();

        RegFile#(Bit#(12), Bit#(8)) mem <- mkRegFileFull();

        rule handleWrite;
            let r <- writeSlave.request.get();
            mem.upd(truncate(tpl_1(r)), tpl_2(r));
        endrule

        rule handleRead;
            let r <- readSlave.request.get();
            readSlave.response.put(mem.sub(truncate(r)));
        endrule

        interface ESABusWr_Slave_Fabric s_wr = writeSlave.fab;
        interface ESABusRd_Slave_Fabric s_rd = readSlave.fab;
    endmodule

    module [BlueCheck] mkESAMemSpec#(Reset r)();
        RegFile#(Bit#(12), Bit#(8)) mem <- mkRegFileFull(reset_by r);
        FIFO#(Bit#(8)) readRequest <- mkFIFO(reset_by r);
        ESAMem impl <- mkESAMem(reset_by r);

        ESABusWr_Master writeMaster <- mkESABusWr_Master(reset_by r);
        ESABusRd_Master readMaster <- mkESABusRd_Master(reset_by r);
        mkConnection(writeMaster.fab, impl.s_wr, reset_by r);
        mkConnection(readMaster.fab, impl.s_rd, reset_by r);

        function Action getMem(Bit#(16) v);
            action
                readRequest.enq(mem.sub(truncate(v)));
            endaction
        endfunction

        function Action writeMem(Tuple2#(Bit#(16), Bit#(8)) r);
            action
                mem.upd(truncate(tpl_1(r)), tpl_2(r));
            endaction
        endfunction

        equiv("upd", writeMem, writeMaster.request.put);
        equiv("sub", getMem, readMaster.request.put);
        equiv("read", toGet(readRequest).get, readMaster.response.get);
    endmodule

    module [Module] mkESAMemChecker ();
        Clock clk <- exposeCurrentClock;
        MakeResetIfc r <- mkReset(0, True, clk);

        // Customise default BlueCheck parameters
        BlueCheck_Params params = bcParamsID(r);
        params.wedgeDetect      = True;
        params.id.initialDepth  = 3;
        function incr(x)        = x+1;
        params.id.incDepth      = incr;
        params.numIterations    = 25;
        params.id.testsPerDepth = 100;

        // Generate checker
        Stmt s <- mkModelChecker(mkESAMemSpec(r.new_rst), params);
        mkAutoFSM(s);
    endmodule

endpackage