package Median;

    import Types :: *;
    import ClientServer :: *;
    import GetPut :: *;
    import Vector :: *;
    import List :: *;
    import FIFO :: *;

    import BlueCheck :: *;

    typedef Vector#(9, GrayScale) FilterKernel;
    typedef Server#(Vector#(9, GrayScale), GrayScale) Median;

    module mkMedian(Median);
        FIFO#(FilterKernel) in <- mkFIFO();
        FIFO#(GrayScale) out <- mkFIFO();

        Vector#(7, Server#(Vector#(3, GrayScale), Sorted)) sortingNetwork <- replicateM(mkSort());

        rule firstStage;
            let t = in.first(); in.deq();

            Vector#(3, Vector#(3,GrayScale)) c = unpack(pack(t));

            for(Integer i = 0; i < 3; i = i + 1) begin
                sortingNetwork[i].request.put(c[i]);
            end
        endrule

        rule secondStage;
            Vector#(3, Vector#(3, GrayScale)) sorted;
            for(Integer i = 0; i < 3; i = i + 1) begin
                let tVal <- sortingNetwork[i].response.get();
                sorted[i] = unpack(pack(tVal));
            end

            for(Integer i = 0; i < 3; i = i + 1) begin
                Vector#(3, GrayScale) iSort;
                for(Integer j = 0; j < 3; j = j + 1) begin
                    iSort[j] = sorted[j][i];
                end
                sortingNetwork[3 + i].request.put(iSort);
            end
        endrule

        rule thirdStage;
            Vector#(3, GrayScale) sorted;
            for(Integer i = 0; i < 3; i = i + 1) begin
                let tVal <- sortingNetwork[3 + i].response.get();
                Vector#(3, GrayScale) tSorted = unpack(pack(tVal));
                sorted[i] = tSorted[2 - i];
            end
            sortingNetwork[6].request.put(sorted);
        endrule

        rule fourthStage;
            let tVal <- sortingNetwork[6].response.get();
            out.enq(tVal.med);
        endrule

        interface Put request = toPut(in);
        interface Get response = toGet(out);
    endmodule

    typedef struct {
        GrayScale max;
        GrayScale med;
        GrayScale min;
    } Sorted deriving(Bits, Eq);

    module mkSort(Server#(Vector#(3, GrayScale), Sorted));
        FIFO#(Vector#(3, GrayScale)) in <- mkFIFO();
        FIFO#(Sorted) out <- mkFIFO();

        rule sort;
            let i = in.first(); in.deq();
            let xored = i[0] ^ i[1] ^ i[2];
            Sorted tVal;
            tVal.max = max(i[0], max(i[1], i[2]));
            tVal.min = min(i[0], min(i[1], i[2]));
            tVal.med = xored ^ tVal.max ^ tVal.min;
            out.enq(tVal);
        endrule

        interface Put request = toPut(in);
        interface Get response = toGet(out);
    endmodule

    module [BlueCheck] mkMedianSpec ();
        Median impl <- mkMedian;

        FIFO#(Vector#(9, GrayScale)) specFIFO <- mkFIFO();

        function ActionValue#(GrayScale) getMedian();
            actionvalue
                let s = specFIFO.first(); specFIFO.deq();
                List#(GrayScale) l = toList(s);
                l = sort(l);

                return l[4];
            endactionvalue
        endfunction

        equiv("put", specFIFO.enq, impl.request.put);
        equiv("get", getMedian, impl.response.get);
    endmodule

    module [Module] mkMedianChecker ();
       blueCheck(mkMedianSpec);
    endmodule

endpackage