package SoC;

import StmtFSM :: *;
import Connectable :: *;
import GetPut :: *;
import ClientServer :: *;
import Vector :: *;

import Types :: *;
import DMA :: *;
import ESABus :: *;
import ESAConnect :: *;
import ESAMem :: *;
import MedianFilter :: *;
import ColorConverter :: *;

module mkSoC(Empty);

ESAMem mem <- mkESAMem();
DMA_MemoryToStream dma_mts <- mkDMA_MemoryToStream();
DMA_StreamToMemory dma_stm <- mkDMA_StreamToMemory();

Server#(GrayScale, GrayScale) median <- mkMedianFilter();
Server#(RGB, GrayScale) converter <- mkGray();

mkConnection(converter.response, median.request);

Reg#(RGB) rgbTmp <- mkRegU;
Reg#(UInt#(2)) rgbCntr <- mkReg(0);
rule byteStreamToRGB;
    let rgb = rgbTmp;
    if(rgbCntr == 0) begin
        rgb.r <- dma_mts.data.get();
        rgbCntr <= 1;
    end else if(rgbCntr == 1) begin
        rgb.g <- dma_mts.data.get();
        rgbCntr <= 2;
    end else begin
        rgb.b <- dma_mts.data.get();
        converter.request.put(rgb);
        rgbCntr <= 0;
    end
    rgbTmp <= rgb;
endrule

mkConnection(median.response, dma_stm.data);

ESABusWr_Master master_write <- mkESABusWr_Master();
ESABusRd_Master master_read <- mkESABusRd_Master();

Vector#(2, ESABusWr_Master_Fabric) master_wr;
master_wr[0] = master_write.fab;
master_wr[1] = dma_stm.m_wr;

Vector#(3, ESABusWr_Slave_Fabric) slaves_wr;
slaves_wr[0] = mem.s_wr;
slaves_wr[1] = dma_mts.s_wr;
slaves_wr[2] = dma_stm.s_wr;

mkESAWrConnector(master_wr, slaves_wr);

Vector#(2, ESABusRd_Master_Fabric) master_rd;
master_rd[0] = master_read.fab;
master_rd[1] = dma_mts.m_rd;

Vector#(3, ESABusRd_Slave_Fabric) slaves_rd;
slaves_rd[0] = mem.s_rd;
slaves_rd[1] = dma_mts.s_rd;
slaves_rd[2] = dma_stm.s_rd;

mkESARdConnector(master_rd, slaves_rd);

Reg#(Bool) mtsInterrupt <- mkReg(False);
rule fetchMTSInterrupt if(dma_mts.doneInterrupt());
    mtsInterrupt <= True;
    $display("(%0d) Memory-Mapped to Stream DMA is done", $time);
endrule

Reg#(Bool) stmInterrupt <- mkReg(False);
rule fetchSTMInterrupt if(dma_stm.doneInterrupt());
    stmInterrupt <= True;
    $display("(%0d) Stream to Memory-Mapped DMA is done", $time);
endrule

Reg#(UInt#(32)) cntr1 <- mkReg(0);
Reg#(UInt#(32)) cntr2 <- mkReg(0);
Vector#(20, Reg#(GrayScale)) tmpRow <- replicateM(mkRegU);
Stmt controller = {
    seq
        $display("(%0d) Write RGB image to memory", $time);
        for(cntr1 <= 0; cntr1 < 20 * 20 * 3; cntr1 <= cntr1 + 1) seq
            master_write.request.put(tuple2(pack(truncate(cntr1)), pack(truncate(cntr1))));
        endseq
        $display("(%0d) Done writing image to memory.", $time);
        $display("(%0d) Setting up and starting DMA engines.", $time);
        // dma_mts starts at 0x4000, Read image from 0
        master_write.request.put(tuple2('h4000, 0));
        master_write.request.put(tuple2('h4001, 0));
        master_write.request.put(tuple2('h4002, 'hb0));
        master_write.request.put(tuple2('h4003, 'h4));
        master_write.request.put(tuple2('h4004, 1));
        // dma_stm starts at 0x8000, Write image starting at 0x800
        master_write.request.put(tuple2('h8000, 'h00));
        master_write.request.put(tuple2('h8001, 'h8));
        master_write.request.put(tuple2('h8002, 'h90));
        master_write.request.put(tuple2('h8003, 'h1));
        master_write.request.put(tuple2('h8004, 1));
        $display("(%0d) Done setting up DMA engines", $time);
        await(mtsInterrupt && stmInterrupt);
        $display("(%0d) Processing done", $time);
        for(cntr1 <= 0; cntr1 < 20; cntr1 <= cntr1 + 1) seq
            for(cntr2 <= 0; cntr2 < 20; cntr2 <= cntr2 + 1) seq
                master_read.request.put(pack(truncate('h12c + (cntr1 * 10) + cntr2)));
                action
                    let r <- master_read.response.get();
                    tmpRow[cntr2] <= r;
                endaction
            endseq
            $display(fshow(readVReg(tmpRow)));
        endseq
    endseq
};

mkAutoFSM(controller);

endmodule

endpackage