package Testbench;

import ColorConverter::*;
import ClientServer::*;
import GetPut :: *;
import Vector :: *;
import FIFO :: *;

typedef 4 TotalTests;

module mkTestbench(Empty);
    Server#(RGB, GrayScale) uut <- mkGray;

    Vector#(TotalTests, Tuple2#(RGB, GrayScale)) testvalues;
    testvalues[0] = tuple2(RGB { r:0, g:0, b:0 }, 0);
    testvalues[1] = tuple2(RGB { r:255, g:255, b:255 }, 254); // Rounding errors/Round towards zero
    testvalues[2] = tuple2(RGB { r:42, g:10, b:50 }, 23); 
    testvalues[3] = tuple2(RGB { r:128, g:5, b:240 }, 68); 

    Reg#(Bool) sendTest <- mkReg(True);
    Reg#(UInt#(32)) testCntr <- mkReg(0);

    rule sendTestValue if(sendTest && testCntr < fromInteger(valueOf(TotalTests)));
        uut.request.put(tpl_1(testvalues[testCntr]));
        sendTest <= False;
    endrule

    rule checkTestValue if(!sendTest && testCntr < fromInteger(valueOf(TotalTests)));
        let r <- uut.response.get();
        if(r != tpl_2(testvalues[testCntr])) begin
            $display("Error at %d: %d != %d", testCntr, r, tpl_2(testvalues[testCntr]));
        end
        testCntr <= testCntr + 1;
        sendTest <= True;
    endrule

    rule endTest if(testCntr == fromInteger(valueOf(TotalTests)));
        $finish();
    endrule
    
endmodule

import BlueCheck :: *;

// Very simple/inefficient implementation
module mkSimpleGray(Server#(RGB, GrayScale));
    FIFO#(GrayScale) outputValue <- mkFIFO;
        
    interface Put request;
        method Action put(RGB rgb);
            UInt#(32) factorR = 76;  // floor(0.299 * (1 << 8))
            UInt#(32) factorG = 150; // floor(0.587 * (1 << 8))
            UInt#(32) factorB = 29;  // floor(0.114 * (1 << 8))

            UInt#(32) r = extend(unpack(rgb.r)) << 8;
            UInt#(32) g = extend(unpack(rgb.g)) << 8;
            UInt#(32) b = extend(unpack(rgb.b)) << 8;

            outputValue.enq(truncate(pack((r * factorR + g * factorG + b * factorB) >> 16)));
        endmethod
    endinterface

    interface Get response = toGet(outputValue);
endmodule

module [BlueCheck] mkColorConverterSpec ();
    Server#(RGB, GrayScale) spec <- mkSimpleGray;
    Server#(RGB, GrayScale) impl <- mkGray();

    equiv("put", spec.request.put, impl.request.put);
    equiv("get", spec.response.get, impl.response.get);
endmodule

module [Module] mkColorConverterChecker ();
    blueCheck(mkColorConverterSpec);
endmodule

endpackage